### CopyBelowMaterials.rb
### by Chanz
require 'sketchup.rb'
###
module DAN
  ###
  def self.copymaterials()
    model=Sketchup.active_model
    ents=model.active_entities
    #ss=model.selection ### use selected upper faces to speed it up
    faces = ents.grep(Sketchup::Face)
    #faces = ss.grep(Sketchup::Face) ### use selected upper faces to speed it up
    ###
    return nil unless faces[0]
    ###
    begin
      model.start_operation("DAN.copymaterials", true)
    rescue
      model.start_operation("DAN.copymaterials")
    end
    ###
    faces.each {|face|
      rayt = model.raytest(self.find_centroid(face), Z_AXIS.reverse)
      if rayt
        rayt[1].each{|f|
          if f.is_a?(Sketchup::Face)
            self.process_face(f, face)
          end
        }
      end
    }
    ###
    model.commit_operation
    ###
  end

  # Finds an adjusted point, so we don't raycast down to an edge :)
  ### OK if triangulated...
  ### BUT IF the face is L shaped the centroid might not fall on it !!!
  def self.find_centroid(face)
    vertices=face.vertices
    vertices[vertices.length]=vertices[0]
    centroid=Geom::Point3d.new()
    a_sum=0.0
    y_sum=0.0
    x_sum=0.0
    for i in (0...vertices.length-1)
      temp=(vertices[i].position.x*vertices[i+1].position.y)-(vertices[i+1].position.x*vertices[i].position.y)
      a_sum+=temp
      x_sum+=(vertices[i+1].position.x+vertices[i].position.x)*temp
      y_sum+=(vertices[i+1].position.y+vertices[i].position.y)*temp
    end###for
    centroid.x=x_sum/(3*a_sum)
    centroid.y=y_sum/(3*a_sum)
    centroid.z=face.bounds.min.z
    return centroid
  end
 
  # NOTE: face_upper is in the model.ents||selection, face_lower is the face the raytest hit
  def self.process_face(face_upper, face_lower)
    ###
    texture_writer=Sketchup.create_texture_writer
    ###
    if face_upper.material && ! face_upper.material.texture
      ### paint the plain material if it exists but it has no texture
      face_lower.material=face_upper.material
    elsif face_upper.material ### A material and it has a texture...
      # Sample the UVs from UPPER face
      samples = []
      samples << face_upper.vertices[0].position             ### 0,0 | Origin
      samples << samples[0].offset(face_upper.normal.axes.x) ### 1,0 | Offset Origin in X
      samples << samples[0].offset(face_upper.normal.axes.y) ### 0,1 | Offset Origin in Y
      samples << samples[1].offset(face_upper.normal.axes.y) ### 1,1 | Offset X in Y
      xyz = [];uv = [] ### Arrays containing 3D and UV points.
      uvh = face_upper.get_UVHelper(true, true, texture_writer)
      samples.each { |position|
        xyz << position ### XYZ 3D coordinates
        uvq = uvh.get_front_UVQ(position) ### UV 2D coordinates
        uv << self.flattenUVQ(uvq)
      }
      pts = [] ### Position texture.
      (0..3).each { |i|
        pts << xyz[i]
        pts << uv[i]
      }
      # Set the position for textured material from face_upper onto face_lower front
      face_lower.position_material(face_upper.material, pts, true)
	  else
	  print "WTF"
    end
    ###
  end
 
  ### Get UV coordinates from UVQ matrix.
  def self.flattenUVQ(uvq)
    return Geom::Point3d.new(uvq.x / uvq.z, uvq.y / uvq.z, 1.0)
  end
 
  ###
  ### menu
  unless file_loaded?(__FILE__)
    menu=UI.menu("Plugins").add_submenu("Copy Materials...")
    menu.add_item("From Nearest Below"){self.copymaterials()}
  end
  file_loaded(__FILE__)

end